import json
import random
import re
from io import BytesIO
from pathlib import Path
from typing import List

import httpx
from loguru import logger
from lxml import etree
from PIL import Image, ImageEnhance

from .model import Operator

resource_path = Path(__file__).parent.parent / "resource"
char_pat = re.compile(r"\|职业=(.+?)\n\|.+?")

six_bgi = Image.open(resource_path / "back_six.png")
five_bgi = Image.open(resource_path / "back_five.png")
four_bgi = Image.open(resource_path / "back_four.png")
low_bgi = Image.new("RGBA", (124, 360), (49, 49, 49))
six_tail = Image.open(resource_path / "six_02.png")

six_line = Image.open(resource_path / "six_01.png")
five_line = Image.open(resource_path / "five.png")
four_line = Image.open(resource_path / "four.png")

star_circle = Image.open(resource_path / "star_02.png")
enhance_five_line = Image.new("RGBA", (124, 720), (0x60, 0x60, 0x60, 0x50))
enhance_four_line = Image.new("RGBA", (124, 720), (132, 108, 210, 0x10))
brighter = ImageEnhance.Brightness(six_line)
six_line = brighter.enhance(1.5)
brighter = ImageEnhance.Brightness(four_line)
four_line = brighter.enhance(0.9)
six_line_up = six_line.crop((0, 0, six_line.size[0], 256))
six_line_down = six_line.crop((0, 256, six_line.size[0], 512))
five_line_up = five_line.crop((0, 0, five_line.size[0], 256))
five_line_down = five_line.crop((0, 256, five_line.size[0], 512))
four_line_up = four_line.crop((0, 0, four_line.size[0], 256))
four_line_down = four_line.crop((0, 256, four_line.size[0], 512))
logger.debug("basic image loaded.")
characters = {
    "先锋": Image.open(resource_path / "图标_职业_先锋_大图_白.png"),
    "近卫": Image.open(resource_path / "图标_职业_近卫_大图_白.png"),
    "医疗": Image.open(resource_path / "图标_职业_医疗_大图_白.png"),
    "术师": Image.open(resource_path / "图标_职业_术师_大图_白.png"),
    "狙击": Image.open(resource_path / "图标_职业_狙击_大图_白.png"),
    "特种": Image.open(resource_path / "图标_职业_特种_大图_白.png"),
    "辅助": Image.open(resource_path / "图标_职业_辅助_大图_白.png"),
    "重装": Image.open(resource_path / "图标_职业_重装_大图_白.png"),
}
logger.debug("careers image loaded.")
stars = {
    5: Image.open(resource_path / "稀有度_白_5.png"),
    4: Image.open(resource_path / "稀有度_白_4.png"),
    3: Image.open(resource_path / "稀有度_白_3.png"),
    2: Image.open(resource_path / "稀有度_白_2.png"),
}
logger.debug("stars image loaded.")

if not (resource_path / "ops_initialized").exists():
    logger.critical("operator resources has not initialized yet")
    exit()

with (resource_path / "careers.json").open("r", encoding="utf-8") as f:
    careers = json.load(f)
operators = {
    path.stem: Image.open(path) for path in (resource_path / "operators").iterdir()
}
logger.debug("operators image loaded.")


async def simulate_image(ops: List[Operator]):
    """
    依据抽卡结果生成模拟十连图片

    :param ops: 抽卡结果
    :return: 图片的bytes
    """
    base = 20
    offset = 124
    l_offset = 14
    back_img = Image.open(resource_path / "back_image.png")
    async with httpx.AsyncClient() as async_httpx:
        for op in ops[:10]:
            name = op.name
            rarity = op.rarity - 1
            try:
                if name in operators and name in careers:
                    avatar: Image.Image = operators[name]
                    logo: Image.Image = characters[careers[name]].resize(
                        (96, 96), Image.Resampling.LANCZOS
                    )
                else:
                    resp = await async_httpx.get(f"https://prts.wiki/w/文件:半身像_{name}_1.png")
                    root = etree.HTML(resp.text)
                    sub = root.xpath(f'//img[@alt="文件:半身像 {name} 1.png"]')[0]
                    avatar: Image.Image = Image.open(
                        BytesIO(
                            (
                                await async_httpx.get(
                                    f"https://prts.wiki{sub.xpath('@src').pop()}"
                                )
                            ).read()
                        )
                    ).crop((20, 0, offset + 20, 360))
                    with (resource_path / "operators" / f"{name}.png").open("wb+") as _f:
                        operators[name] = avatar
                        avatar.save(
                            _f, format="PNG", quality=100, subsampling=2, qtables="web_high"
                        )
                    resp1 = await async_httpx.get(
                        f"https://prts.wiki/index.php?title={name}&action=edit"
                    )
                    root1 = etree.HTML(resp1.text)
                    sub1 = root1.xpath('//textarea[@id="wpTextbox1"]')[0]
                    cr = char_pat.search(sub1.text)[1]
                    logo: Image.Image = characters[cr].resize(
                        (96, 96), Image.Resampling.LANCZOS
                    )
                    with (resource_path / "careers.json").open("w", encoding="utf-8") as jf:
                        careers[name] = cr
                        json.dump(careers, jf, ensure_ascii=False)
            except (ValueError, IndexError):
                resp = await async_httpx.get("https://prts.wiki/w/文件:半身像_无_1.png")
                root = etree.HTML(resp.text)
                sub = root.xpath('//img[@alt="文件:半身像 无 1.png"]')[0]
                logo: Image.Image = characters[random.choice(list(characters))].resize(
                    (96, 96), Image.Resampling.LANCZOS
                )
                avatar: Image.Image = Image.open(
                    BytesIO(
                        (
                            await async_httpx.get(
                                f"https://prts.wiki{sub.xpath('@src').pop()}"
                            )
                        ).read()
                    )
                ).resize((offset, 360), Image.Resampling.LANCZOS)

            s_size = stars[rarity].size
            star = stars[rarity].resize(
                (int(s_size[0] * 0.6), int(47 * 0.6)), Image.Resampling.LANCZOS
            )
            s_offset = (offset - int(star.size[0])) // 2

            if rarity == 5:
                back_img.paste(six_line_up, (base, 0), six_line_up)
                back_img.paste(six_line_down, (base, 720 - 256), six_line_down)
                back_img.paste(six_tail, (base, 0), six_tail)
                back_img.paste(
                    six_tail.transpose(Image.Transpose.ROTATE_180),
                    (base, 720 - 256),
                    six_tail.transpose(Image.Transpose.ROTATE_180),
                )
                basei = six_bgi.copy()
            elif rarity == 4:
                back_img.paste(enhance_five_line, (base, 0), enhance_five_line)
                back_img.paste(five_line_up, (base, 0), five_line_up)
                back_img.paste(five_line_down, (base, 720 - 256), five_line_down)
                basei = five_bgi.copy()
            elif rarity == 3:
                back_img.paste(enhance_four_line, (base, 0), enhance_four_line)
                back_img.paste(four_line_up, (base, 0), four_line_up)
                back_img.paste(four_line_down, (base, 720 - 256), four_line_down)
                back_img.paste(star_circle, (base - 2, 180 - 64), star_circle)
                basei = four_bgi.copy()
            else:
                basei = low_bgi.copy()
            size = avatar.size
            avatar.thumbnail(size)
            basei.paste(avatar, (0, 0), avatar)
            back_img.paste(basei, (base, 180))
            s_size = star.size
            star.thumbnail(s_size)
            back_img.paste(star, (base + s_offset, 166), star)
            l_size = logo.size
            logo.thumbnail(l_size)
            back_img.paste(logo, (base + l_offset, 492), logo)
            base += offset
        imageio = BytesIO()
        back_img.save(
            imageio,
            format="PNG",
            quality=80,
            subsampling=2,
            qtables="web_high",
        )
        return imageio.getvalue()



import json
from datetime import datetime
from pathlib import Path
from random import Random
from typing import List, Optional, TypedDict


class Tags(TypedDict):
    static: str
    optional: List[str]


class SubCareer(TypedDict):
    name: str
    block: int
    attack_speed: str
    cost_basic: int
    talent: str
    tags: Tags


class Career(TypedDict):
    name: str
    info: List[SubCareer]


class Information(TypedDict):
    career: List[Career]
    organize: List[str]
    homeland: List[str]
    race: List[str]
    phy_exam_evaul: List[str]


class RandomOperator:
    """依据名称随机生成干员"""
    rand_operator_dict: Information

    def __init__(self, path: Optional[str] = None):
        """
        :param path: 模板文件路径
        """
        path = (
            Path(path)
            if path
            else Path(__file__).parent / "resource" / "operator_template.json"
        )
        with path.open("r+", encoding="utf-8") as f:
            self.rand_operator_dict = json.load(f)

    def generate(self, name: str) -> str:
        """
        依据名称随机生成干员
        """
        rand = Random()
        count = sum(ord(char) for char in name)
        now = datetime.now()
        rand.seed(count + (now.day * now.month) + now.year)

        career_dict = rand.choice(self.rand_operator_dict["career"])
        career_name = career_dict["name"]
        career_info_dict = rand.choice(career_dict["info"])

        sex = "男" if rand.randint(1, 2) == 1 else "女"
        career_detail = career_info_dict["name"]
        level = rand.randint(3, 6)
        cost = (
                career_info_dict["cost_basic"]
                + level
                - (2 if rand.randint(0, 4) >= 2 else rand.randint(0, 1))
        )
        attack_speed = career_info_dict["attack_speed"]
        block = career_info_dict["block"]
        talent = career_info_dict["talent"]
        tags = career_info_dict["tags"]["static"] + rand.choice(
            career_info_dict["tags"]["optional"]
        )
        infect = "参照医学检测报告，确认为" + ("感染者。" if rand.randint(0, 10) > 5 else "非感染者。")
        race = rand.choice(self.rand_operator_dict["race"])
        homeland = rand.choice(self.rand_operator_dict["homeland"])
        organize = rand.choice(self.rand_operator_dict["organize"])
        if organize.endswith("homeland"):
            organize = rand.choice(self.rand_operator_dict["homeland"])
        height = rand.randint(0, 25) + (165 if sex == "男" else 155)
        fight_exp = (
            rand.randint(0, 6) if rand.randint(0, 10) > 2 else rand.randint(6, 20)
        )

        return "\n".join(
            [
                f"{name}",
                f"{'★' * level}",
                f"【性别】{sex}",
                f"【职业】{career_name}-{career_detail}",
                f"【初始费用】{cost}",
                f"【阻挡数】{block}",
                f"【攻击速度】{attack_speed}",
                f"【特性】{talent}",
                f"【标签】{tags}",
                "\n",
                f"【种族】{race}",
                f"【身高】{height} cm",
                f"【出生地】{homeland}",
                f"【所属阵营】{organize}",
                f"【战斗经验】{'无' if fight_exp == 0 else f'{fight_exp}年'}" "\n",
                f"【矿石病感染情况】{infect}",
                f"【物理强度】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
                f"【战场机动】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
                f"【生理耐受】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
                f"【战术规划】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
                f"【战斗技巧】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
                f"【源石技艺适应性】{rand.choice(self.rand_operator_dict['phy_exam_evaul'])}",
            ]
        )



import itertools
import json
import math
import random
from io import BytesIO
from pathlib import Path
from typing import List, Optional, Union

from loguru import logger
from PIL import Image, ImageDraw, ImageFont

from ..util import random_pick_big
from .model import GachaData, GachaUser, Operator
from .update import update
from .download import download


class ArknightsGacha:
    """抽卡模拟器"""
    five_per: int
    four_per: int
    three_per: int
    data: GachaData

    color = {
        6: (0xFF, 0x7F, 0x27),  # ff7f27
        5: (0xFF, 0xC9, 0x0E),  # ffc90e
        4: (0x93, 0x19, 0x93),  # d8b3d8
        3: (0x09, 0xB3, 0xF7),  # 09b3f7
    }

    def __init__(self, file: Optional[Union[str, Path]] = None):
        """
        :param file: 卡池信息文件
        """
        self.five_per, self.four_per, self.three_per = 8, 50, 40
        if not file:
            file = Path(__file__).parent.parent / "resource" / "default_gacha.json"
        elif isinstance(file, str):
            file = Path(file)
            if not file.exists():
                raise FileNotFoundError
        self.file = file
        with file.open("r", encoding="UTF-8") as f_obj:
            self.data = json.load(f_obj)

    async def initialize(self, retry: int = 5):
        if (self.file.parent / "ops_initialized").exists():
            return
        ops = self.data["operators"]["三"] + self.data["operators"]["四"] + self.data["operators"]["五"] + self.data["operators"]["六"]
        await download(ops, retry)
        logger.success("operator resources initialized")

    async def update(self):
        """更新当前卡池"""
        resp = await update()
        if not resp:
            return
        if resp.title == self.data["name"]:
            return
        if self.data["name"] != "常驻标准寻访" and self.data["six_per"] < 1:
            self.data["operators"]["六"] += self.data["up_six_list"]
            self.data["operators"]["六"] = list(set(self.data["operators"]["六"]))
            self.data["operators"]["五"] += self.data["up_five_list"]
            self.data["operators"]["五"] = list(set(self.data["operators"]["五"]))
            self.data["operators"]["四"] += self.data["up_four_list"]
            self.data["operators"]["四"] = list(set(self.data["operators"]["四"]))
        self.data["up_six_list"].clear()
        self.data["up_five_list"].clear()
        self.data["up_four_list"].clear()
        self.data["up_limit"].clear()
        self.data["up_alert_limit"].clear()
        self.data["name"] = resp.title
        for char in resp.six_chars:
            if char.limit:
                if not self.data["up_limit"]:
                    self.data["up_limit"].append(char.name)
                else:
                    self.data["up_alert_limit"].append(char.name)
                    continue
            else:
                self.data["up_six_list"].append(char.name)
            self.data["six_per"] = char.chance
        for char in resp.five_chars:
            self.data["up_five_list"].append(char.name)
            self.data["five_per"] = char.chance
        for char in resp.four_chars:
            self.data["up_four_list"].append(char.name)
            self.data["four_per"] = char.chance

        with self.file.open("w", encoding="UTF-8") as f_obj:
            json.dump(self.data, f_obj, ensure_ascii=False, indent=4)
        return resp

    def gacha(self, user: GachaUser, count: int = 1) -> List[List[Operator]]:
        """
        模拟抽卡，返回生成的干员信息

        :param user: 抽卡用户，用来继承抽卡概率
        :param count: 抽卡数量
        :return: 干员信息
        """
        gacha_ranks: List[List[Operator]] = []
        cache = []
        for i in range(1, count + 1):
            x = random_pick_big(
                "六五四三", [user.six_per, self.five_per, self.four_per, self.three_per]
            )
            ans = "".join(itertools.islice(x, 1))
            if ans != "六":
                user.six_statis += 1
                if user.six_statis > 50:
                    user.six_per += 2
                    if self.three_per > 1:
                        self.three_per -= 2
                    elif self.four_per > 1:
                        self.four_per -= 2
                    elif self.five_per > 1:
                        self.five_per -= 2
            else:
                user.six_statis = 0
                user.six_per = 2
                self.five_per, self.four_per, self.three_per = 8, 50, 40

            cache.append(self.generate_operator(ans))
            if i % 10 == 0:
                gacha_ranks.append(cache)
                cache = []
        if cache:
            gacha_ranks.append(cache)
        return gacha_ranks

    def generate_operator(self, rank: str) -> Operator:
        """
        抽取单个干员

        :param rank: 干员等级，从 六、五、四、三中选取
        :return: 生成的干员信息
        """
        card_list = self.data["operators"][rank].copy()
        if rank == "六":
            if (six_per := self.data["six_per"]) >= 1.0:
                return Operator(random.choice(self.data["up_six_list"]), 6)
            up_res = random.choice(self.data["up_six_list"] + self.data["up_limit"])
            for c in self.data["up_alert_limit"]:
                card_list.extend([c for _ in range(5)])
            card_list.extend(
                [up_res for _ in range(int(len(card_list) * six_per / (1 - six_per)))]
            )
            random.shuffle(card_list)
            return Operator(random.choice(card_list), 6)
        if rank == "五":
            if (five_per := self.data["five_per"]) >= 1.0:
                return Operator(random.choice(self.data["up_five_list"]), 5)
            up_res = random.choice(self.data["up_five_list"])
            card_list.extend(
                [up_res for _ in range(int(len(card_list) * five_per / (1 - five_per)))]
            )
            random.shuffle(card_list)
            return Operator(random.choice(card_list), 5)
        if rank == "四":
            if self.data["up_four_list"]:
                four_per = self.data["four_per"]
                up_res = random.choice(self.data["up_four_list"])
                card_list.extend(
                    [
                        up_res
                        for _ in range(int(len(card_list) * four_per / (1 - four_per)))
                    ]
                )
            return Operator(random.choice(card_list), 4)
        return Operator(random.choice(card_list), 3)

    def create_image(
        self,
        user: GachaUser,
        result: List[List[Operator]],
        count: int = 1,
        relief: bool = False,
    ) -> bytes:
        """
        将抽卡结果转为图片

        :param user: 抽卡用户
        :param result: 本次抽卡结果
        :param count: 抽卡数量
        :param relief: 是否需要浮雕效果
        :return: 生成的图片bytes
        """
        tile = 20
        width_base = 720
        color_base = 0x40
        color_bases = (color_base, color_base, color_base)
        height = tile * int(math.ceil(count / 10) + 1) + 130
        img = Image.new("RGB", (width_base, height), color_bases)
        # 绘画对象
        draw = ImageDraw.Draw(img)
        font_base = ImageFont.truetype("simhei.ttf", 16)
        draw.text(
            (tile, tile), "博士小心地拉开了包的拉链...会是什么呢？", fill="lightgrey", font=font_base
        )

        pool = f"当前卡池:【{self.data['name']}】"
        draw.text(
            (width_base - font_base.getsize(pool)[0] - tile, tile),
            pool,
            fill="lightgrey",
            font=font_base,
        )
        if relief:
            xi = 2 * tile
            yi = 2 * tile + 4
            xj = width_base - (2 * tile)
            yj = tile * (int(math.ceil(count / 10)) + 4)
            for i in range(3, 0, -1):
                d = int(color_base * 0.2) // 4
                r = int(color_base * 0.8) + i * d
                draw.rounded_rectangle(
                    (xi - i, yi - i, xi + i, yj + i), radius=16, fill=(r, r, r)
                )
                draw.rounded_rectangle(
                    (xj - i, yi - i, xj + i, yj + i), radius=16, fill=(r, r, r)
                )
            for i in range(4, 0, -1):
                r = (color_base // 4) * i
                draw.rounded_rectangle(
                    (xi - i, yi - i, xj + i, yi + i),
                    radius=20,
                    fill=(r, r, r, int(256 * 0.6)),
                )
                d = (0xFF - color_base) // 4
                r = 0xFF - i * d
                draw.rounded_rectangle(
                    (xi - i, yj - i, xj + i, yj + i),
                    radius=20,
                    fill=(r, r, r, int(256 * 0.8)),
                )
        for i, ots in enumerate(result):
            base = tile * 3
            if relief:
                draw.rounded_rectangle(
                    (
                        base,
                        tile * (i + 3) + 4,
                        base + tile * 3 * len(ots) - 2,
                        tile * (i + 4) + 3,
                    ),
                    radius=2,
                    fill=(color_base // 2, color_base // 2, color_base // 2),
                )
            for operator in ots:
                width = tile * 3
                length = len(operator.name)
                length = max(length, 3)
                font_size = int(3 * font_base.size / length)
                font = font_base.font_variant(size=font_size)
                width_offset = (width - font.getsize(operator.name)[0]) // 2
                height_offset = 1 + (tile - font.getsize(operator.name)[1]) // 2

                draw.rounded_rectangle(
                    (base, tile * (i + 3) + 2, base + width - 2, tile * (i + 4)),
                    radius=2,
                    fill=self.color[operator.rarity],
                )
                draw.text(
                    (base + width_offset, tile * (i + 3) + height_offset),
                    operator.name,
                    fill="#ffffff",
                    stroke_width=1,
                    stroke_fill=tuple(
                        int(i * 0.5) for i in self.color[operator.rarity]
                    ),
                    font=font,
                )
                base += width
        draw.text(
            (tile, height - 3 * tile + 10),
            f"博士已经抽取了{user.six_statis}次没有6星了" f"\n当前出6星的机率为 {user.six_per}%",
            fill="lightgrey",
            font=font_base,
        )
        imageio = BytesIO()
        img.save(
            imageio,
            format="PNG",
            quality=90,
            subsampling=2,
            qtables="web_high",
        )
        return imageio.getvalue()

    def gacha_with_img(self, user: GachaUser, count: int = 1, relief: bool = True):
        return self.create_image(user, (self.gacha(user, count)), count, relief)



from typing import TypedDict, Dict, List, NamedTuple
from dataclasses import dataclass, field


class Operator(NamedTuple):
    name: str
    rarity: int


class GachaData(TypedDict):
    name: str
    six_per: float
    five_per: float
    four_per: float
    operators: Dict[str, List[str]]  # 当期新 up 的干员不应该在里面
    up_limit: List[str]
    up_alert_limit: List[str]
    up_six_list: List[str]
    up_five_list: List[str]
    up_four_list: List[str]


@dataclass
class GachaUser:
    six_per: int = field(default=2)
    six_statis: int = field(default=0)


class UpdateChar(NamedTuple):
    name: str
    limit: bool
    chance: float


@dataclass
class UpdateResponse:
    title: str
    four_chars: List[UpdateChar]
    five_chars: List[UpdateChar]
    six_chars: List[UpdateChar]
    pool: str



import asyncio
import json
import re
from io import BytesIO
from pathlib import Path
from typing import List

import httpx
from loguru import logger
from lxml import etree
from PIL import Image

char_pat = re.compile(r"\|职业=(.+?)\n\|.+?")

career = {}
base_path = Path(__file__).parent.parent / "resource" / "operators"
base_path.mkdir(parents=True, exist_ok=True)
if (base_path.parent / "careers.json").exists():
    with (base_path.parent / "careers.json").open("r", encoding="utf-8") as f:
        career.update(json.load(f))


async def download(ops: List[str], retry: int = 5):
    async with httpx.AsyncClient() as async_httpx:
        for name in ops:
            if (base_path / f"{name}.png").exists():
                continue
            while retry:
                try:
                    resp = await async_httpx.get(
                        f"https://prts.wiki/w/文件:半身像_{name}_1.png", timeout=20.0
                    )
                    root = etree.HTML(resp.text)
                    sub = root.xpath(f'//img[@alt="文件:半身像 {name} 1.png"]')[0]
                    avatar: Image.Image = Image.open(
                        BytesIO(
                            (
                                await async_httpx.get(
                                    f"https://prts.wiki{sub.xpath('@src').pop()}"
                                )
                            ).read()
                        )
                    ).crop((20, 0, 124 + 20, 360))
                    with (base_path / f"{name}.png").open("wb+") as img:
                        avatar.save(
                            img,
                            format="PNG",
                            quality=100,
                            subsampling=2,
                            qtables="web_high",
                        )
                    resp1 = await async_httpx.get(
                        f"https://prts.wiki/index.php?title={name}&action=edit"
                    )
                    root1 = etree.HTML(resp1.text)
                    sub1 = root1.xpath('//textarea[@id="wpTextbox1"]')[0]
                    career[name] = char_pat.search(sub1.text)[1]
                    logger.success(f"{name}({career[name]}) saved")
                    break
                except Exception as e:
                    retry -= 1
                    logger.warning(f"Fetch {name} failed. Caused by {e}")
                    logger.debug("Retrying...")
                    await asyncio.sleep(2)
            if not retry:
                with (base_path.parent / "careers.json").open(
                    "w+", encoding="utf-8"
                ) as f:
                    json.dump(career, f, ensure_ascii=False)
                logger.critical(f"Fetch {name} failed. Exiting...")
                return
            await asyncio.sleep(3)
        with (base_path.parent / "careers.json").open("w+", encoding="utf-8") as f:
            json.dump(career, f, ensure_ascii=False)
        with (base_path.parent / "ops_initialized").open("w+", encoding="utf-8") as f:
            f.write("initialized")



import json
import random
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Dict, List, Literal, Optional, TypedDict, overload
import re
import httpx
from lxml import etree
from PIL import Image, ImageDraw, ImageFont

__all__ = ["Operator", "Guess", "GuessUnit", "OperatorWordle", "update"]

rarity_pat = re.compile(r"\|稀有度=(\d+?)\n\|.+?")
char_pat = re.compile(r"\|职业=([^|]+?)\n\|.+?")
sub_char_pat = re.compile(r"\|分支=([^|]+?)\n\|.+?")
race_pat = re.compile(r"\|种族=([^|]+?)\n\|.+?")
org_pat = re.compile(r"\|所属国家=([^|]+?)\n\|.+?")
org_pat1 = re.compile(r"\|所属组织=([^|]+?)\n\|.+?")
org_pat2 = re.compile(r"\|所属团队=([^|]+?)\n\|.+?")
art_pat = re.compile(r"\|画师=([^|]+?)\n\|.+?")
base = Path(__file__).parent / "resource" / "wordle"
sign = {
    "correct": Image.open(base / "correct.png"),
    "down": Image.open(base / "down.png"),
    "up": Image.open(base / "up.png"),
    "wrong": Image.open(base / "wrong.png"),
    "relate": Image.open(base / "relate.png"),
}
simple_sign = {"correct": "🟩", "down": "🟦", "up": "🟦", "wrong": "🟥", "relate": "🟨"}

state = Literal["correct", "down", "up", "wrong", "relate"]


class Operator(TypedDict):
    rarity: int
    org: str
    career: str
    race: str
    artist: str
    relate: Optional[List[str]]


class GuessUnit(TypedDict):
    rarity: state
    org: state
    career: state
    race: state
    artist: state
    name: str


@dataclass
class Guess:
    state: Literal["success", "guessing", "failed"]
    lines: List[GuessUnit]
    select: str


with (base / "relations.json").open("r", encoding="utf-8") as f:
    _data = json.load(f)
    relations: Dict[str, List[str]] = _data["org_related"]
    tables: Dict[str, Operator] = _data["table"]


def update(*names: str):
    for name in names:
        resp = httpx.get(f"https://prts.wiki/index.php?title={name}&action=edit")
        root = etree.HTML(resp.text, etree.HTMLParser())
        sub = root.xpath('//textarea[@id="wpTextbox1"]')[0].text
        char = char_pat.search(sub)[1]
        sub_char = sub_char_pat.search(sub)[1]
        rarity = rarity_pat.search(sub)[1]
        try:
            race = race_pat.search(sub)[1]
        except TypeError:
            race = "/"
        try:
            org1 = org_pat.search(sub)[1]
        except TypeError:
            org1 = ""
        try:
            org2 = org_pat1.search(sub)[1]
        except TypeError:
            org2 = ""
        try:
            org3 = org_pat2.search(sub)[1]
        except TypeError:
            org3 = ""
        org = org3 if org3 else org2 if org2 else org1
        org = org or "/"
        art = art_pat.search(sub)[1]
        tables[name] = {
            "rarity": int(rarity),
            "org": org,
            "career": f"{char}-{sub_char}",
            "race": race,
            "artist": art,
        }
    with (base / "relations.json").open("w", encoding="utf-8") as _f:
        json.dump(_data, _f, ensure_ascii=False)


class OperatorWordle:
    def __init__(self, path: str, max_guess: int = 8):
        self.base_dir = Path(path)
        self.base_dir.mkdir(parents=True, exist_ok=True)
        if not self.base_dir.is_dir():
            raise NotADirectoryError(path)
        self.max_guess = max_guess

    def restart(self, uid: str):
        data_path = self.base_dir / f"{uid}.json"
        data_path.unlink(missing_ok=True)

    def select(self, uid: str):
        selected_name = random.choice(list(tables.keys()))
        selected = tables[selected_name]
        with (self.base_dir / f"{uid}.json").open("w+", encoding="utf-8") as _f:
            json.dump(
                {
                    "select": selected,
                    "select_name": selected_name,
                    "select_time": 0,
                    "units": [],
                },
                _f,
                ensure_ascii=False,
                indent=2,
            )
        return selected_name, selected

    def guess(self, name: str, uid: str) -> Guess:
        data_path = self.base_dir / f"{uid}.json"
        if not data_path.exists():
            old_res = []
            select_time = 0
            selected_name, selected = self.select(uid)
        else:
            with data_path.open("r", encoding="utf-8") as _f:
                sdata = json.load(_f)
                selected_name = sdata["select_name"]
                selected = sdata["select"]
                old_res = sdata["units"]
                select_time = sdata["select_time"]
        if name not in tables:
            raise ValueError("干员不存在")
        res = {
            "rarity": "correct",
            "org": "correct",
            "career": "correct",
            "race": "correct",
            "artist": "correct",
            "name": name,
        }
        if name == selected_name:
            data_path.unlink()
            return Guess("success", old_res + [res], selected_name)
        select_time += 1
        guess_op = tables[name]
        if guess_op["rarity"] < selected["rarity"]:
            res["rarity"] = "up"
        elif guess_op["rarity"] > selected["rarity"]:
            res["rarity"] = "down"

        if guess_op["org"] != selected["org"]:
            if selected_name in guess_op.get("relate", []) or name in selected.get(
                "relate", []
            ):
                res["org"] = "relate"
            elif guess_op["org"] in relations[selected["org"]]:
                res["org"] = "relate"
            else:
                res["org"] = "wrong"

        if guess_op["career"] != selected["career"]:
            g_cs = guess_op["career"].split("-")
            res["career"] = (
                "relate" if selected["career"].startswith(g_cs[0]) else "wrong"
            )

        if guess_op["race"] != selected["race"]:
            res["race"] = "wrong"

        if guess_op["artist"] != selected["artist"]:
            res["artist"] = "wrong"
        if select_time >= self.max_guess:
            data_path.unlink()
            return Guess("failed", old_res + [res], selected_name)
        with data_path.open("w+", encoding="utf-8") as _f:
            json.dump(
                {
                    "select": selected,
                    "select_name": selected_name,
                    "select_time": select_time,
                    "units": old_res + [res],
                },
                _f,
                ensure_ascii=False,
                indent=2,
            )
        return Guess("guessing", old_res + [res], selected_name)

    @overload
    def draw(self, res: Guess, simple=True) -> str:
        ...

    @overload
    def draw(self, res: Guess, simple=False) -> bytes:
        ...

    def draw(self, res: Guess, simple: bool = False):
        if simple:
            ans = f"干员猜猜乐 {len(res.lines)}/{self.max_guess}\n"
            for unit in res.lines:
                ans += simple_sign[unit["rarity"]]
                ans += simple_sign[unit["org"]]
                ans += simple_sign[unit["career"]]
                ans += simple_sign[unit["race"]]
                ans += simple_sign[unit["artist"]]
                ans += unit["name"]
                ans += "\n"
            return ans

        back_img = Image.new("RGB", (600, 80 * (len(res.lines) + 2)), (0, 0, 0))
        draw = ImageDraw.Draw(back_img)
        font_base = ImageFont.truetype("simhei.ttf", 20)
        draw.text((20, 45), "稀有度", fill="white", font=font_base)
        draw.text((130, 45), "阵营", fill="white", font=font_base)
        draw.text((230, 45), "职业", fill="white", font=font_base)
        draw.text((330, 45), "种族", fill="white", font=font_base)
        draw.text((430, 45), "画师", fill="white", font=font_base)
        draw.text((530, 45), "名称", fill="white", font=font_base)
        for index, unit in enumerate(res.lines):
            slot = sign[unit["rarity"]].copy()
            size = slot.size
            slot.thumbnail(size)
            back_img.paste(slot, (10, 80 * (index + 1)), slot)
            slot = sign[unit["org"]].copy()
            size = slot.size
            slot.thumbnail(size)
            back_img.paste(slot, (110, 80 * (index + 1)), slot)
            slot = sign[unit["career"]].copy()
            size = slot.size
            slot.thumbnail(size)
            back_img.paste(slot, (210, 80 * (index + 1)), slot)
            slot = sign[unit["race"]].copy()
            size = slot.size
            slot.thumbnail(size)
            back_img.paste(slot, (310, 80 * (index + 1)), slot)
            slot = sign[unit["artist"]].copy()
            size = slot.size
            slot.thumbnail(size)
            back_img.paste(slot, (410, 80 * (index + 1)), slot)
            length = len(unit["name"])
            length = max(length, 4)
            font_size = int(4 * font_base.size / length)
            font = font_base.font_variant(size=font_size)
            width_offset = (100 - font.getbbox(unit["name"])[2]) // 2
            height_offset = (80 - font.getbbox(unit["name"])[3]) // 2
            draw.text(
                (500 + width_offset, 80 * (index + 1) + height_offset),
                unit["name"],
                fill="pink",
                font=font,
            )
        if res.state == "failed":
            text = f"失败了！这只神秘的干员是{res.select}！"
        elif res.state == "success":
            text = f"成功了！这只神秘的干员是{res.select}！"
        else:
            text = f"你有{len(res.lines)}/{self.max_guess}次机会猜测这只神秘干员，试试看！"
        width = font_base.getbbox(text)
        draw.text(
            ((600 - width[2]) // 2, 80 * (len(res.lines) + 1) + 30),
            text,
            fill="white",
            font=font_base,
        )
        imageio = BytesIO()
        back_img.save(
            imageio,
            format="PNG",
            quality=90,
            subsampling=2,
            qtables="web_high",
        )
        return imageio.getvalue()


if __name__ == "__main__":
    guess = OperatorWordle("./")
    while True:
        name = input(">>> ")
        my_res = guess.guess(name, "aaa")
        print(guess.draw(my_res, simple=True))
        if my_res.state != "guessing":
            break



"""从 prts 查询并截图"""

import playwright.async_api
from typing import Optional
import re

url = "https://prts.wiki/w/%s"
m_url = "https://m.prts.wiki/w/%s"

__all__ = [
    "query",
    "full_page",
    "information",
    "talent",
    "feature",
    "document",
    "upgrade",
    "latent",
    "skills",
    "skill_at",
    "logistics",
    "attributes",
    "attack_range",
    "material",
]


async def full_page(page: playwright.async_api.Page, name: str):
    """全身图"""
    await page.click("html")
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.locator('span[role="button"]:has-text("技能")').first.click()
    await page.wait_for_timeout(2000)
    await page.locator('span[role="button"]:has-text("技能")').first.click()
    await page.locator('span[role="button"]:has-text("精英化材料")').click()
    await page.wait_for_timeout(1000)
    await page.locator('span[role="button"]:has-text("精英化材料")').click()
    await page.locator('span[role="button"]:has-text("相关道具")').click()
    await page.wait_for_timeout(1000)
    await page.locator('span[role="button"]:has-text("相关道具")').click()
    await page.locator('span[role="button"]:has-text("干员档案")').click()
    await page.wait_for_timeout(1000)
    return await page.screenshot(full_page=True)


async def information(
        page: playwright.async_api.Page, name: str, upgraded: bool = False
):
    await page.goto(url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    but = page.locator("text=ELITE2精英二 >> img")
    if (await but.count()) and upgraded:
        await but.first.click()
        await page.locator(
            "text=个人工具 中文（中国大陆）‎创建账号登录 命名空间 页面讨论 变体 查看 阅读查看源代码查看历史 更多 搜索 搜索 前往"
        ).click()
    elem = page.locator("#charimg-mask")
    bounding = await elem.bounding_box()
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def feature(page: playwright.async_api.Page, name: str):
    await page.goto(url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("特性")')
    elem = page.locator('table:has-text("分支 描述")')
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding["height"] += bounding1["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def attributes(page: playwright.async_api.Page, name: str):
    await page.goto(url % name, timeout=600000, wait_until="domcontentloaded")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("属性")')
    elem = page.locator("//table[@class='wikitable logo char-extra-attr-table']")
    elem1 = page.locator("//table[@class='wikitable logo char-base-attr-table']")
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding2 = await elem1.bounding_box()
    bounding["height"] += bounding1["height"] + 14
    bounding["height"] += bounding2["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def attack_range(page: playwright.async_api.Page, name: str):
    await page.goto(url % name, timeout=600000, wait_until="domcontentloaded")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("攻击范围")')
    elem = page.locator('table:has-text("精英0 精英1 精英2")').first
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding["height"] += bounding1["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def talent(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="domcontentloaded")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("天赋")')
    block = page.locator(
        "//section[@class='mf-section-6 collapsible-block open-block']"
    )
    elem = block.locator("//table[@class='wikitable logo']").first
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding2 = await block.bounding_box()
    bounding["height"] += bounding2["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def latent(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="domcontentloaded")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("潜能提升")')
    block = page.locator("//section[@id='content-collapsible-block-6']")
    elem = block.locator("//table[@class='wikitable nodesktop logo']").first
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding2 = await block.bounding_box()
    bounding["height"] += bounding2["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def skills(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("技能")').first
    block = page.locator("//section[@id='content-collapsible-block-7']")
    if not (await block.locator("//table[@class='wikitable nodesktop logo']").count()):
        bounding = await title.bounding_box()
        bounding1 = await block.bounding_box()
        bounding["height"] += bounding1["height"] + 14
    else:
        elem = block.locator("//table[@class='wikitable nodesktop logo']").first
        await elem.click()
        await page.wait_for_timeout(100)
        await page.click('[aria-label="用户导航"]')
        bounding = await title.bounding_box()
        bounding1 = await elem.bounding_box()
        bounding2 = await block.bounding_box()
        bounding["height"] += bounding2["height"] + 14
        bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def skill_at(page: playwright.async_api.Page, name: str, index: int = 0):
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    block = page.locator("//section[@id='content-collapsible-block-7']")
    skill = block.locator("//table[@class='wikitable nodesktop logo']")
    count = await skill.count()
    if not count or index > count:
        title = page.locator('h2:has-text("技能")').first
        bounding = await title.bounding_box()
        bounding1 = await block.bounding_box()
        bounding["height"] += bounding1["height"] + 14
        return [await page.screenshot(full_page=True, clip=bounding, scale="device")]
    res = []
    for i in range(count):
        await skill.nth(i).click()
        await page.wait_for_timeout(500)
        await page.click('[aria-label="用户导航"]')
        bounding = await skill.nth(i).bounding_box()
        res.append(await page.screenshot(full_page=True, clip=bounding, scale="device"))
    if count < 1:
        return res
    return res[count - 1: count]


async def logistics(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("后勤技能")').first
    block = page.locator("//section[@id='content-collapsible-block-8']")
    elem = block.locator("//table[@class='wikitable logo']").first
    await elem.click()
    await page.wait_for_timeout(1000)
    await page.click('[aria-label="用户导航"]')
    bounding = await title.bounding_box()
    bounding1 = await elem.bounding_box()
    bounding2 = await block.bounding_box()
    bounding["height"] += bounding2["height"] + 28
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def upgrade(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("精英化材料")').first
    if await page.locator('h2:has-text("召唤物信息")').count():
        block = page.locator("//section[@id='content-collapsible-block-10']")
    else:
        block = page.locator("//section[@id='content-collapsible-block-9']")
    if not (await block.locator("//table[@class='wikitable logo']").count()):
        bounding = await title.bounding_box()
        bounding1 = await block.bounding_box()
        bounding["height"] += bounding1["height"] + 14
    else:
        elem = block.locator("//table[@class='wikitable logo']").first
        await elem.click()
        await page.wait_for_timeout(1000)
        await page.click('[aria-label="用户导航"]')
        bounding = await title.bounding_box()
        bounding1 = await elem.bounding_box()
        bounding2 = await block.bounding_box()
        bounding["height"] += bounding2["height"] + 28
        bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def material(page: playwright.async_api.Page, name: str):
    await page.goto(url % name, timeout=600000, wait_until="load")
    await page.wait_for_timeout(100)
    title = page.locator('h2:has-text("技能升级材料")')
    elem = page.locator('table:has-text("技能升级 1→2")')
    bounding = await title.bounding_box()
    if not (await elem.count()):
        bounding1 = await page.locator("text=该干员没有技能").nth(1).bounding_box()
    else:
        await title.click()
        await page.wait_for_timeout(2000)
        await page.click(f'h1:has-text("{name}")')
        bounding1 = await elem.first.bounding_box()
    bounding["height"] += bounding1["height"] + 14
    bounding["width"] = bounding1["width"]
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


async def document(page: playwright.async_api.Page, name: str):
    await page.goto(m_url % name, timeout=600000, wait_until="load")
    await page.locator("span[role=\"button\"]:has-text(\"展开\")").first.click()
    await page.click('[aria-label="用户导航"]')
    elem = page.locator(
        "//table[@class='wikitable mw-collapsible logo mw-made-collapsible']"
    )
    bounding = await elem.bounding_box()
    return await page.screenshot(full_page=True, clip=bounding, scale="device")


_table = {
    "((干员)?信息|展示)": information,
    "(模板|职业)?特性": feature,
    "属性": attributes,
    "攻击范围": attack_range,
    "天赋": talent,
    "潜能(提升)?": latent,
    "技能(信息)?": skills,
    "后勤技能": logistics,
    "精(一|二)": upgrade,
    "精英化(材料)?": upgrade,
    "技能升级材料": material,
    "(.*?)专(一|二|三|精)(材料)?": material,
    "档案": document
}


def _handle(content: Optional[str] = None):
    if not content:
        return full_page
    for pat, func in _table.items():
        if re.fullmatch(pat, content):
            return func
    return full_page


async def query(
        name: str,
        content: Optional[str] = None,
        page: Optional[playwright.async_api.Page] = None,
):
    if not page:
        async with playwright.async_api.async_playwright() as pw:
            browser = await pw.chromium.launch(headless=True)
            context = await browser.new_context(viewport={"width": 1440, "height": 960})
            page = await context.new_page()

            func = _handle(content)
            try:
                data = await func(page, name)
            except playwright.async_api.TimeoutError:
                data = await full_page(page, name)
            finally:
                await page.close()
                await context.close()
                await browser.close()
            return data
    func = _handle(content)
    try:
        return await func(page, name)
    except playwright.async_api.TimeoutError:
        return await full_page(page, name)

if __name__ == "__main__":
    import asyncio


    async def main():
        data = await query("艾雅法拉", "档案")
        with open("temp.png", "wb+") as f:
            f.write(data)


    asyncio.run(main())



"""
Copy from nonebot_plugin_gamedraw
"""
import re
from typing import List

import httpx
from loguru import logger
from lxml import etree

from .model import UpdateChar, UpdateResponse


async def update():
    """
    从方舟官网处获取卡池更新信息

    :return: 更新结果
    """
    async with httpx.AsyncClient() as async_httpx:
        result = (await async_httpx.get("https://ak.hypergryph.com/news.html")).text
        if not result:
            logger.warning("明日方舟 获取公告出错")
            return
        dom = etree.HTML(result, etree.HTMLParser())
        activity_urls = dom.xpath(
            "//ol[@class='articleList' and @data-category-key='ACTIVITY']/li/a/@href"
        )
        up_chars: List[List[UpdateChar]] = [[], [], []]
        pool_img = ""
        for activity_url in activity_urls[:20]:  # 减少响应时间, 10个就够了
            activity_url = f"https://ak.hypergryph.com{activity_url}"
            result = (await async_httpx.get(activity_url)).text
            if not result:
                logger.warning(f"明日方舟 获取公告 {activity_url} 出错")
                continue

            """因为鹰角的前端太自由了，这里重写了匹配规则以尽可能避免因为前端乱七八糟而导致的重载失败"""
            dom = etree.HTML(result, etree.HTMLParser())
            contents = dom.xpath(
                "//div[@class='article-content']/p/text() | //div[@class='article-content']/p/span/text() | //div["
                "@class='article-content']/div[@class='media-wrap image-wrap']/img/@src "
            )
            title = ""
            chars: List[str] = []
            for index, content in enumerate(contents):
                if not re.search("(.*)(寻访|复刻).*?开启", content):
                    continue
                title = re.split(r"[【】]", content)
                title = "".join(title[1:-1]) if "-" in title else title[1]
                lines = [contents[index + _] for _ in range(8)] + [""]
                for idx, line in enumerate(lines):
                    """因为 <p> 的诡异排版，所以有了下面的一段"""
                    if "★★" in line and "%" in line:
                        chars.append(line)
                    elif "★★" in line and "%" in lines[idx + 1]:
                        chars.append(line + lines[idx + 1])
                pool_img = contents[index - 2]
                r"""两类格式：用/分割，用\分割；★+(概率)+名字，★+名字+(概率)"""
                for char in chars:
                    star = char.split("（")[0].count("★")
                    name = (
                        re.split(r"[（：]", char)[1]
                        if "★（" not in char
                        else re.split("（|）：", char)[2]
                    )
                    names = name.replace("\\", "/").split("/")
                    for name in names:
                        limit = False
                        if "[限定]" in name:
                            limit = True
                        name = name.replace("[限定]", "").strip()
                        zoom = 1.0
                        if match := re.search(r"（在.*?以.*?(\d+).*?倍.*?）", char) or re.search(
                                r"（占.*?的.*?(\d+).*?%）", char
                        ):
                            zoom = float(match[1])
                            zoom = zoom / 100 if zoom > 10 else zoom
                        up_chars[6 - star].append(UpdateChar(name, limit, zoom))
                break  # 这里break会导致个问题：如果一个公告里有两个池子，会漏掉下面的池子，比如 5.19 的定向寻访。但目前我也没啥好想法解决
            if not title:
                continue
            logger.debug(f"成功获取 当前up信息; 当前up池: {title}")
            return UpdateResponse(title, up_chars[2], up_chars[1], up_chars[0], pool_img)



from pathlib import Path

from loguru import logger

from .main import ArknightsGacha
from .model import GachaUser

if not (Path(__file__).parent.parent / "resource" / "ops_initialized").exists():
    logger.critical("Operator Resources has not initialized yet")
    logger.warning("Please execute ArknightsGacha.initialize()")
    logger.warning("Otherwise, you can't use simulate-image function")
else:
    from .simulate import simulate_image



import random
from typing import List, TypeVar, Iterable

T = TypeVar("T")


def random_pick_big(sequence: Iterable[T], relative_odds: List[int]):
    """
    sequence: [a, b, c, d]
    relative_odds: [pa, pb, pc, pd], 其中 pa~pd > 0
    """
    table = [z for x, y in zip(sequence, relative_odds) for z in [x] * y]
    while True:
        yield random.choice(table)


def random_pick_small(some_list: List[T], probabilities: List[float]) -> T:
    """
    some_list: [a, b, c, d]
    probabilities: [pa, pb, pc, pd], 其中 0 < pa ~ pd < 1
    """
    x = random.uniform(0, 1)
    cumulative_probability = 0.0
    item = some_list[0]
    for item, item_probability in zip(some_list, probabilities):
        cumulative_probability += item_probability
        if x < cumulative_probability:
            break
    return item
